/*
 * Licensed Materials - Property of IBM
 *
 * com.ibm.etools.c
 *
 *  Copyright IBM Corporation 2004. All Rights Reserved. 
 *
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.etools.c.importer.AST;

import java.util.ArrayList;
import java.util.Iterator;

import org.eclipse.cdt.core.parser.ast.ASTClassKind;
import org.eclipse.cdt.core.parser.ast.ASTPointerOperator;
import org.eclipse.cdt.core.parser.ast.IASTAbstractDeclaration;
import org.eclipse.cdt.core.parser.ast.IASTArrayModifier;
import org.eclipse.cdt.core.parser.ast.IASTClassSpecifier;
import org.eclipse.cdt.core.parser.ast.IASTExpression;
import org.eclipse.cdt.core.parser.ast.IASTField;
import org.eclipse.cdt.core.parser.ast.IASTSimpleTypeSpecifier;
import org.eclipse.cdt.core.parser.ast.IASTSimpleTypeSpecifier.Type;
import org.eclipse.cdt.cpp.miners.parser.expressionevaluator.ExpressionEvaluator;
import org.eclipse.cdt.dstore.core.model.DataElement;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTClassSpecifier;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTElaboratedTypeSpecifier;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTEnumerationSpecifier;
import org.eclipse.cdt.internal.core.parser.ast.complete.ASTTypedef;

import com.ibm.etools.c.CArray;
import com.ibm.etools.c.CDatatype;
import com.ibm.etools.c.CDerivableType;
import com.ibm.etools.c.CFactory;
import com.ibm.etools.c.CPointer;
import com.ibm.etools.c.CStruct;
import com.ibm.etools.c.CTypedef;
import com.ibm.etools.c.CUnion;
import com.ibm.etools.c.datatypes.CBitField;
import com.ibm.etools.c.datatypes.CEnumeration;
import com.ibm.etools.c.datatypes.CIntegral;
import com.ibm.etools.c.datatypes.DatatypesFactory;
import com.ibm.etools.c.importer.CException;
import com.ibm.etools.c.importer.CFunctionParser;
import com.ibm.etools.c.importer.CIM;
import com.ibm.etools.c.importer.CModelConverter;
import com.ibm.etools.c.importer.CPreferenceStore;
import com.ibm.etools.c.importer.CPrimitiveTypesMapper;
import com.ibm.etools.c.importer.CResource;
import com.ibm.etools.c.importer.CTypeTable;
import com.ibm.etools.c.importer.CUtility;

/**
 * The CTypeParser class parses the string representation of a type and create a
 * corresponding C model type object.
 * 
 * The exposed interfaces are the two findOrCreateType() methods, which are used
 * to map a string to a C model type represented by a CClassifier object.
 * 
 * The created types are saved in a CTypeTable, so that we only have to create
 * each type once. The next time we can retrieve it from the table.
 */
public class CTypeParser {

    public static final String copyright = "Licensed Material - Property of IBM\n com.ibm.etools.c\n(C) Copyright IBM Corp. 2004 - All Rights Reserved.\nUS Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp."; //$NON-NLS-1$

    private CTypeTable _typeTable;

    private CModelConverter _modelConverter;

    // A factory used to generate C language entities.
    private CFactory _factory;

    // A factory used to generate C datatypes.
    private DatatypesFactory _datatypesFactory;

    /**
     * Construct a CTypeParser from a CModelConverter.
     */
    public CTypeParser(CModelConverter modelConverter) {

        _modelConverter = modelConverter;
        _typeTable = modelConverter.getTypeTable();
        _factory = CFactory.eINSTANCE;
        _datatypesFactory = DatatypesFactory.eINSTANCE;
    }

    /**
     * Find or create a type from a given string. If the type already exists, we
     * simply return it. Otherwise we call createType() to create a new type
     * from the string.
     * <p>
     * This is the exposed interface that the user should call.
     */
    public CDerivableType findOrCreateType(String typeString, boolean arraySizeCanBeZero) throws Exception {

        CDerivableType existingType = _typeTable.getType(typeString);

        if (existingType == null) {
            existingType = createType(typeString, arraySizeCanBeZero);
        }

        return existingType;
    }

    /**
     * Find or create a type from a given string. If the type already exists, we
     * simply return it. Otherwise we call createType() to create a new type
     * from the string.
     * <p>
     * This is the exposed interface that the user should call.
     */
    public CDerivableType findOrCreateType(IASTAbstractDeclaration iASTAbstractDeclaration, String typeString, boolean arraySizeCanBeZero) throws Exception {

        Iterator arrayModifiersIterator = iASTAbstractDeclaration.getArrayModifiers();

        Iterator pointerOperators = iASTAbstractDeclaration.getPointerOperators();

        if (pointerOperators != null) {
            while (pointerOperators.hasNext()) {
                ASTPointerOperator aSTPointerOperator = (ASTPointerOperator) pointerOperators.next();
                typeString += " *";
            }
        }

        while (arrayModifiersIterator.hasNext()) {
            IASTArrayModifier iASTArrayModifier = (IASTArrayModifier) arrayModifiersIterator.next();
            IASTExpression iASTExpression = iASTArrayModifier.getExpression();
            typeString += "[" + iASTExpression.toString() + "]";
        }

        CDerivableType existingType = _typeTable.getType(typeString);

        if (existingType != null) {
            return existingType;
        } else {
            return createType(iASTAbstractDeclaration, typeString, arraySizeCanBeZero);
        }
    }

    /**
     * Create a type from the given type string This is an internal method.
     */
    private CDerivableType createType(String typeString, boolean arraySizeCanBeZero) throws Exception {

        // Return null if the type string is empty
        // Defect24937: Added the third condition to this if check, since
        // pointers to array
        // confuse the processing...for example int (*ptr)[10]; gets converted
        // to (*)[10] ptr
        // by typeListToString(), and then treated as a function pointer.
        if (typeString == null || typeString.length() == 0 || (typeString.indexOf("(*)[") >= 0))
            return null;

        CDerivableType returnType = null;

        // First check if this is a function pointer
        if (isFunctionPointer(typeString)) {
            CFunctionParser fpt = new CFunctionParser(typeString, _modelConverter);
            returnType = fpt.getPointer();
            return returnType;
        } else if (isFunction(typeString)) {
            CFunctionParser fp = new CFunctionParser(typeString, _modelConverter);
            returnType = fp.getFunction();
            return returnType;
        }

        char lastChar = typeString.charAt(typeString.length() - 1);

        // Is it a pointer?
        if (lastChar == CIM.P_CHAR_POINTER) {
            returnType = createPointer(typeString);
        }
        // Is it an array?
        else if (lastChar == CIM.P_CHAR_RIGHT_BRACKET) {
            returnType = createArray(typeString, arraySizeCanBeZero);
        }
        // Is it a bitfield?
        else if (isBitField(typeString)) {
            returnType = createBitField(typeString);
        }
        // Is it a C primitive type?
        else if (isPrimitiveType(typeString)) {
            CDatatype primType = CPrimitiveTypesMapper.getTypeFromName(typeString);
            _typeTable.addType(typeString, primType);
            returnType = primType;
        }
        // When we get here, it means that the type is unknown at the moment.
        // Depending on user settings, we either throw an exception or create a
        // user-defined type.
        else if (CPreferenceStore.isTreatUnknownTypeAsError()) {
            throw new CException(CResource.getString("_ERROR_Unknown_Type", typeString));
        }
        // It is a user defined type if it does not satisfy any of the above
        else {
            returnType = createUserType(typeString);
        }

        return returnType;
    }

    /**
     * Create a type from the given type string This is an internal method.
     */
    private CDerivableType createType(IASTAbstractDeclaration iASTAbstractDeclaration, String typeString, boolean arraySizeCanBeZero) throws Exception {

        // Return null if the type string is empty
        // Defect24937: Added the third condition to this if check, since
        // pointers to array
        // confuse the processing...for example int (*ptr)[10]; gets converted
        // to (*)[10] ptr
        // by typeListToString(), and then treated as a function pointer.
        if (typeString == null || typeString.length() == 0 || (typeString.indexOf("(*)[") >= 0))
            return null;

        CDerivableType returnType = null;

        // First check if this is a function pointer
        if (isFunctionPointer(typeString)) {
            CFunctionParser fpt = new CFunctionParser(typeString, _modelConverter);
            returnType = fpt.getPointer();
            return returnType;
        } else if (isFunction(typeString)) {
            CFunctionParser fp = new CFunctionParser(typeString, _modelConverter);
            returnType = fp.getFunction();
            return returnType;
        }

        char lastChar = typeString.charAt(typeString.length() - 1);

        // Is it a pointer?
        if (lastChar == CIM.P_CHAR_POINTER) {
            returnType = createPointer(typeString);
        }
        // Is it an array?
        else if (lastChar == CIM.P_CHAR_RIGHT_BRACKET) {
            returnType = createArray(typeString, arraySizeCanBeZero);
        }
        // Is it a bitfield?
        else if (isBitField(typeString)) {
            returnType = createBitField(typeString);
        }
        // Is it a C primitive type?
        else if (isPrimitiveType(typeString)) {
            CDatatype primType = CPrimitiveTypesMapper.getTypeFromName(typeString);
            _typeTable.addType(typeString, primType);
            returnType = primType;
        }
        // When we get here, it means that the type is unknown at the moment.
        // Depending on user settings, we either throw an exception or create a
        // user-defined type.
        else if (CPreferenceStore.isTreatUnknownTypeAsError()) {
            throw new CException(CResource.getString("_ERROR_Unknown_Type", typeString));
        }
        // It is a user defined type if it does not satisfy any of the above
        else {
            returnType = createUserType(typeString);
        }

        return returnType;
    }

    public CDerivableType findOrCreateType(IASTField typedElement, String structName, String unionName, boolean arraySizeCanBeZero) throws Exception {

        String typeString = "";
        String name = "";

        IASTAbstractDeclaration iASTAbstractDeclaration = typedElement.getAbstractDeclaration();
        Object typeSpecifierObject = iASTAbstractDeclaration.getTypeSpecifier();

        if (typeSpecifierObject instanceof IASTSimpleTypeSpecifier) {
            IASTSimpleTypeSpecifier iASTSimpleTypeSpecifier = (IASTSimpleTypeSpecifier) typeSpecifierObject;

            if (iASTSimpleTypeSpecifier.getType() == Type.CLASS_OR_TYPENAME) {
                Object object = iASTSimpleTypeSpecifier.getTypeSpecifier();

                if (object instanceof ASTEnumerationSpecifier) {
                    typeString += "enum ";

                } else if (object instanceof ASTClassSpecifier) {
                    ASTClassSpecifier aSTClassSpecifier = (ASTClassSpecifier) object;

                    if (aSTClassSpecifier.getClassKind() == ASTClassKind.STRUCT) {
                        typeString = "struct ";

                    } else if (aSTClassSpecifier.getClassKind() == ASTClassKind.ENUM) {
                        typeString = "enum ";

                    } else if (aSTClassSpecifier.getClassKind() == ASTClassKind.UNION) {
                        typeString = "union ";

                    }
                }
            }

            typeString += iASTSimpleTypeSpecifier.getTypename();

            /*
             * check if it is a BIT field
             */
            if (typedElement.getBitfieldExpression() != null) {
                if (typedElement.getBitfieldExpression().getExpressionKind() == IASTExpression.Kind.PRIMARY_INTEGER_LITERAL) {
                    typeString += ":" + typedElement.getBitfieldExpression().getLiteralString();
                }
            }

        } else if (typeSpecifierObject instanceof IASTClassSpecifier) {
            IASTClassSpecifier iASTClassSpecifier = (IASTClassSpecifier) typeSpecifierObject;

            if (iASTClassSpecifier.getClassKind() == ASTClassKind.STRUCT) {
                name = iASTClassSpecifier.getName();
                
                if (name == null || name.length() <= 0) {
                    name = new String(structName);
                }
                
                typeString = "struct " + name;

            } else if (iASTClassSpecifier.getClassKind() == ASTClassKind.ENUM) {
                typeString = "enum " + iASTClassSpecifier.getName();

            } else if (iASTClassSpecifier.getClassKind() == ASTClassKind.UNION) {
                name = iASTClassSpecifier.getName();
                
                if (name == null || name.length() <= 0) {
                    name = new String(unionName);
                }
                
                typeString = "union " + name;

            }

        } else if (typeSpecifierObject instanceof ASTEnumerationSpecifier) {
            typeString += "enum " + ((ASTEnumerationSpecifier) typeSpecifierObject).getName();

        } else if (typeSpecifierObject instanceof ASTElaboratedTypeSpecifier) {
            ASTElaboratedTypeSpecifier aSTElaboratedTypeSpecifier = (ASTElaboratedTypeSpecifier) typeSpecifierObject;

            if (aSTElaboratedTypeSpecifier.getClassKind() == ASTClassKind.STRUCT) {
                name = aSTElaboratedTypeSpecifier.getName();
                
                if (name == null || name.length() <= 0) {
                    name = new String(structName);
                }
                
                typeString = "struct " + name;

            } else if (aSTElaboratedTypeSpecifier.getClassKind() == ASTClassKind.ENUM) {
                typeString = "enum " + aSTElaboratedTypeSpecifier.getName();

            } else if (aSTElaboratedTypeSpecifier.getClassKind() == ASTClassKind.UNION) {
                name = aSTElaboratedTypeSpecifier.getName();
                
                if (name == null || name.length() <= 0) {
                    name = new String(unionName);
                }
                
                typeString = "union " + name;

            }
        }

        return findOrCreateType(iASTAbstractDeclaration, typeString, arraySizeCanBeZero);
    }

    public CDerivableType findOrCreateType(ASTTypedef typedElement, String typeName, boolean arraySizeCanBeZero) throws Exception {

        String typeString = null;

        IASTAbstractDeclaration iASTAbstractDeclaration = typedElement.getAbstractDeclarator();

        Object typeSpecifierObject = typedElement.getFinalTypeSpecifier();

        if (typeSpecifierObject instanceof IASTSimpleTypeSpecifier) {
            IASTSimpleTypeSpecifier iASTSimpleTypeSpecifier = (IASTSimpleTypeSpecifier) typeSpecifierObject;
            typeString = iASTSimpleTypeSpecifier.getTypename();

        } else if (typeSpecifierObject instanceof IASTClassSpecifier) {
            IASTClassSpecifier iASTClassSpecifier = (IASTClassSpecifier) typeSpecifierObject;

            String name = iASTClassSpecifier.getName();

            if (name == null || name.length() <= 0) {
                name = new String(typeName);
            }

            if (iASTClassSpecifier.getClassKind() == ASTClassKind.STRUCT) {
                typeString = "struct " + name;

            } else if (iASTClassSpecifier.getClassKind() == ASTClassKind.ENUM) {
                typeString = "enum " + name;

            } else if (iASTClassSpecifier.getClassKind() == ASTClassKind.UNION) {
                typeString = "union " + name;

            }
        } else if (typeSpecifierObject instanceof ASTEnumerationSpecifier) {
            ASTEnumerationSpecifier aSTEnumerationSpecifier = (ASTEnumerationSpecifier) typeSpecifierObject;

            String name = aSTEnumerationSpecifier.getName();

            if (name == null || name.length() <= 0) {
                name = new String(typeName);
            }

            typeString = "enum " + name;

        } else if (typeSpecifierObject instanceof ASTElaboratedTypeSpecifier) {
            ASTElaboratedTypeSpecifier aSTElaboratedTypeSpecifier = (ASTElaboratedTypeSpecifier) typeSpecifierObject;

            String name = aSTElaboratedTypeSpecifier.getName();

            if (name == null || name.length() <= 0) {
                name = new String(typeName);
            }

            if (aSTElaboratedTypeSpecifier.getClassKind() == ASTClassKind.STRUCT) {
                typeString = "struct " + name;

            } else if (aSTElaboratedTypeSpecifier.getClassKind() == ASTClassKind.ENUM) {
                typeString = "enum " + name;

            } else if (aSTElaboratedTypeSpecifier.getClassKind() == ASTClassKind.UNION) {
                typeString = "union " + name;

            }
        }

        if (iASTAbstractDeclaration != null) {
            return findOrCreateType(iASTAbstractDeclaration, typeString, arraySizeCanBeZero);

        }

        return findOrCreateType(typeString, arraySizeCanBeZero);
    }

    /**
     * Convert a DataElement array which represents a type to a string
     */
    public String typeListToString(ArrayList typeList) {

        StringBuffer buffer = new StringBuffer();
        boolean bufferIsEmpty = true;
        for (int i = 0; i < typeList.size(); i++) {
            DataElement childElement = (DataElement) typeList.get(i);
            String childName = childElement.getName();
            String childType = childElement.getType();
            if (childName.equals(CIM.P_STR_POINTER)) {
                buffer.append(CIM.P_CHAR_POINTER);
            } else if (childName.charAt(0) == CIM.P_CHAR_LEFT_BRACKET) {
                buffer.append(childName);
            } else if (childType.equals(CIM.P_FUNCTION_TYPE) || childType.equals(CIM.P_MAIN_FUNCTION_TYPE)) {
                buffer.append("[F]" + CUtility.normalizedType(childElement.getValue()));
            }
            // Do not add the const, volatile, inline, static modifiers to the
            // type list
            else if (!childName.equals("const") && !childName.equals("volatile") && !childName.equals("static") && !childName.equals("inline")) {
                if (!bufferIsEmpty) {
                    buffer.append(CIM.P_CHAR_SPACE);
                }
                buffer.append(childName);
                bufferIsEmpty = false;
            }
        }
        return buffer.toString().trim();
    }

    /**
     * Create a pointer from a type string
     */
    private CPointer createPointer(String typeString) throws Exception {

        int length = typeString.length();
        String baseTypeString = typeString.substring(0, length - 1).trim();
        CDerivableType baseType = findOrCreateType(baseTypeString, false);
        CPointer pointer = _factory.createCPointer();
        pointer.setDerives((CDerivableType) baseType);
        _typeTable.addType(typeString, pointer);

        return pointer;
    }

    /**
     * Create an array from a type string
     */
    private CArray createArray(String typeString, boolean arraySizeCanBeZero) throws Exception {

        int openBracketIndex = typeString.indexOf(CIM.P_CHAR_LEFT_BRACKET);

        if (openBracketIndex < 1) {
            throw new CException(CResource.getString("_ERROR_Open_Bracket_Not_Found", typeString));
        }

        int closeBracketIndex = CUtility.findMatchingBracket(typeString, openBracketIndex, true);

        if (closeBracketIndex < 1) {
            throw new CException(CResource.getString("_ERROR_Close_Bracket_Not_Found", typeString));
        }

        String baseTypeString = typeString.substring(0, openBracketIndex) + typeString.substring(closeBracketIndex + 1);
        ;
        String sizeString = typeString.substring(openBracketIndex + 1, closeBracketIndex);

        if (sizeString == null || sizeString.length() == 0) {
            if (arraySizeCanBeZero)
                sizeString = "0";
            else
                throw new CException(CResource.getString("_ERROR_Array_Size_Not_Found", typeString));
        }

        int size = 0;

        try {
            size = Integer.decode(sizeString).intValue();
        } catch (Exception e) {
            // If the sizeString is a constant expression (e.g. (3+2)*5), the
            // parseInt method
            // will throw an exception. Use the expression evaluator to get the
            // array size
            // in this case.
            //
            ExpressionEvaluator eval = new ExpressionEvaluator();
            size = eval.evaluate(sizeString).intValue();
        }

        CDerivableType baseType = findOrCreateType(baseTypeString, arraySizeCanBeZero);
        CArray array = _factory.createCArray();
        array.setDerives(baseType);
        array.setDimension(new Integer(size));
        _typeTable.addType(typeString, array);

        return array;
    }

    /**
     * Create a user defined type (struct, union or typedef) from a given type
     * string
     */
    private CDerivableType createUserType(String typeString) throws Exception {

        if (typeString.startsWith(CIM.P_STRUCT_TYPE)) {
            return createStruct(typeString);
        } else if (typeString.startsWith(CIM.P_UNION_TYPE)) {
            return createUnion(typeString);
        } else if (typeString.startsWith(CIM.P_ENUMERATION_TYPE)) {
            return createEnum(typeString);
        } else {
            return createTypedef(typeString);
        }
    }

    /**
     * Create a structure from a type string
     */
    private CStruct createStruct(String typeString) throws Exception {

        String structName = typeString.substring(7).trim();
        CStruct astruct = _factory.createCStruct();
        astruct.setName(structName);
        _typeTable.addType(typeString, astruct);
        return astruct;
    }

    /**
     * Create a union from a type string
     */
    private CUnion createUnion(String typeString) throws Exception {

        String unionName = typeString.substring(6).trim();
        CUnion aunion = _factory.createCUnion();
        aunion.setName(unionName);
        _typeTable.addType(typeString, aunion);
        return aunion;

    }

    /**
     * Create an enum from a type string
     */
    private CEnumeration createEnum(String typeString) throws Exception {

        String enumName = typeString.substring(5).trim();
        CEnumeration anenum = _datatypesFactory.createCEnumeration();
        anenum.setName(enumName);
        _typeTable.addType(typeString, anenum);
        return anenum;

    }

    /**
     * Create a typedef from a type string
     */
    private CTypedef createTypedef(String typeString) throws Exception {

        CTypedef atypedef = _factory.createCTypedef();
        atypedef.setName(typeString);
        _typeTable.addType(typeString, atypedef);
        return atypedef;
    }

    /**
     * Is it an integral type?
     */
    private boolean isIntegral(String typeString) {

        return (typeString.equals("int") || typeString.equals("short") || typeString.equals("long") || typeString.equals("char") || typeString.equals("unsigned") || typeString.equals("unsigned int")
                || typeString.equals("unsigned short") || typeString.equals("unsigned long") || typeString.equals("short int") || typeString.equals("long int") || typeString.equals("unsigned char")
                || typeString.equals("unsigned short int") || typeString.equals("unsigned long int") || typeString.equals("signed") || typeString.equals("signed int")
                || typeString.equals("signed short") || typeString.equals("signed long") || typeString.equals("signed char") || typeString.equals("signed short int")
                || typeString.equals("signed long int") || typeString.equals("int signed") || typeString.equals("int unsigned") || typeString.equals("short signed")
                || typeString.equals("short unsigned") || typeString.equals("long signed") || typeString.equals("long unsigned") || typeString.equals("char signed")
                || typeString.equals("char unsigned") || typeString.equals("short int signed") || typeString.equals("short int unsigned") || typeString.equals("short signed int")
                || typeString.equals("short unsigned int") || typeString.equals("long int signed") || typeString.equals("long int unsigned") || typeString.equals("long signed int") || typeString
                .equals("long unsigned int"));

    }

    /**
     * Does this type string represent a bitfield?
     */
    private boolean isBitField(String typeString) {
        // If the type string contains the ':' character, then treat it as
        // a bitfield.
        return (typeString.indexOf(CIM.P_CHAR_BITFIELD) > 0);
    }

    /**
     * Does this type string represent a primitive type?
     */
    private boolean isPrimitiveType(String typeString) {
        return (CPrimitiveTypesMapper.getTypeFromName(typeString) != null);
    }

    /**
     * Create a bitfield from a type string
     */
    private CBitField createBitField(String typeString) throws Exception {

        int index = typeString.indexOf(CIM.P_CHAR_BITFIELD);
        String sizeString = typeString.substring(index + 1).trim();
        int size = 0;

        try {
            size = Integer.parseInt(sizeString);
        } catch (Exception e) {
            ExpressionEvaluator eval = new ExpressionEvaluator();
            size = eval.evaluate(sizeString).intValue();
        }

        String baseTypeString = typeString.substring(0, index).trim();
        if (isIntegral(baseTypeString)) {
            CBitField abitfield = _datatypesFactory.createCBitField();
            CIntegral baseType = (CIntegral) CPrimitiveTypesMapper.getTypeFromName(baseTypeString);

            if (baseType != null)
                abitfield.setBaseType(baseType);

            abitfield.setSize(new Integer(size));
            _typeTable.addType(typeString, abitfield);
            return abitfield;
        } else {
            throw new CException(CResource.getString("_ERROR_Bitfield_Basetype_Is_Not_Integral", typeString));
        }
    }

    /**
     * Does this type string represent a function pointer?
     */
    private boolean isFunctionPointer(String typeString) {

        // Treat the type string as a function pointer if it contains the
        // function pointer
        // signature "(*)".
        if (typeString.indexOf(CIM.P_FUNCTION_PTR) >= 0) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Does this type string represent a function?
     */
    private boolean isFunction(String typeString) {

        // Treat the type string as a function if it contains the signature
        // signature "()(".
        if (typeString.indexOf(CIM.P_FUNCTION_SIGNATURE) >= 0) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Does the DataElement representing a nested type contribute to the size of
     * the outer struct/union? A nested type has size effect if it has no
     * following declaration variables.
     * 
     * Example: "struct B" contributes to the size of "struct A" in the
     * following example. struct A { struct B { int i; }; float f; };
     */
    public boolean nestedTypeHasSizeEffect(IASTClassSpecifier element) {
        boolean hasSizeEffect = true;
        return hasSizeEffect;
    }

}